import logging
import time
import typing as t
from types import NoneType

import requests
from urllib3.exceptions import TimeoutError

from . import __version__
from .model import *
from .model.basket import Basket
from .model.payment import PaymentGetResponse, PaymentRequest, PaymentResponse
from .model.paymentpage import PaymentPage, PaymentPageResponse
from .model.webhook import Webhook

logger = logging.getLogger("unzer-sdk")

HttpMethod = t.Literal["GET", "POST", "PUT", "PATCH", "DELETE"]


class UnzerClient:
    endpoint = "https://api.unzer.com/v1"
    retryDelays = (1, 2, 4, 8)
    timeout = 5

    def __init__(
            self,
            private_key: str,
            public_key: str,
            sandbox: bool = False,
            language: str = "en",
    ):
        super(UnzerClient, self).__init__()
        self.private_key = private_key
        self.public_key = public_key
        self.sandbox = sandbox
        self.language = language

    def request(self, operation: str, method: HttpMethod, payload: t.Any = None) -> t.Any:
        """Perform a request to the unzer-api.

        This method does not really perform the request itself,
        but rather prepares the request for :meth:`_request`.

        :param operation: The HTTP method (e.g. POST, GET).
        :param method: The method on the REST API. (path).
        :param payload: The payload for this request.
            Send json-encoded as body.
        :return: The json-decoded response from the api.
        """
        url = "%s/%s" % (self.endpoint, operation)
        headers = {
            "user-agent": "unzer-python-sdk %s" % __version__,
            "content-type": "application/json; charset=UTF-8",
            "accept": "application/json",
            "accept-language": self.language,  # language for translation of customerMessage in errors
        }
        return self._request(
            url,
            method,
            headers,
            payload,
            auth=(self.private_key, "")
        )

    def _request(self, url: str, method: str,
                 headers: list[tuple] | dict[str, str], payload: t.Any,
                 auth: tuple[str, str]) -> t.Any:
        """Helper method to perform the request with throttling.

        :param url: The complete URL.
        :param method: The HTTP method (e.g. POST, GET).
        :param headers: The HTTP headers.
        :type headers: list[tuple] | dict[str, str]
        :param payload: The HTTP payload (will be json encoded).
        :param auth: The authentication for this request.
        :return: The json decoded response

        :raises: :exc:`ErrorResponse` in case of an client error
            or after last retry failed.
        """
        r = None
        for idx, delay in enumerate((0,) + self.retryDelays):
            logger.debug("Perform try no. %d (delay: %d)", idx, delay)
            time.sleep(delay)
            logger.debug("%s %s", method, url)
            logger.debug("payload: %r", payload)
            try:
                r = requests.request(
                    method,
                    url,
                    json=payload,
                    headers=headers,
                    auth=auth,
                    verify=True,
                    timeout=self.timeout,
                )
            except (TimeoutError, requests.exceptions.ReadTimeout):
                logger.exception("Caught TimeoutError")
                continue
            if 200 <= r.status_code <= 201:
                logger.debug("Response[%s %s]: %r", r.status_code, r.reason, r.json())
                return r.json()
            elif 500 <= r.status_code < 600:
                logger.debug("Server error")
                logger.debug("Response[%s %s]: %r", r.status_code, r.reason, r.text)
                continue
            else:
                logger.debug("Client error")
                logger.debug("Response[%s %s]: %r", r.status_code, r.reason, r.text)
                errorResponse = ErrorResponse.fromDict(r.json())
                errorResponse.statusCode = r.status_code
                errorResponse.srcResponse = r
                raise errorResponse

        logger.error("All request attempts failed")
        if r is not None:
            try:
                errorResponse = ErrorResponse.fromDict(r.json(), "All request attempts failed")
                errorResponse.statusCode = r.status_code
                errorResponse.srcResponse = r
            except ValueError:
                logger.exception("Failed to build an ErrorResponse from last request")
            else:
                raise errorResponse
        raise ErrorResponse("All request attempts failed", srcResponse=r)

    def getKeyPair(self) -> dict:
        """Provides the public key of the used private key as well as a list of the payment types available for the merchant.

        :return: The fetched KeyPairResponse
        """
        # ToDo: implement KeyPairResponse-model
        return self.request(
            "keypair",
            "GET",
        )

    def getKeyPairTypes(self) -> dict[str, t.Any]:
        """Get detailed information and payment method configuration

        The endpoint provides details about the key pairs configured for a merchant account
        and the payment methods they support, including their associated public key, private key,
        currencies, and customer types (like B2C or B2B).

        .. seealso::

            - https://api.unzer.com/api-reference/index.html#tag/Keypair/operation/getAvailablePaymentMethodTypesWithTypeInformation  # noqa: E501
            - https://docs.unzer.com/server-side-integration/direct-api-integration/manage-api-resources/api-check-key-configuration/  # noqa: E501

        :return: The fetched KeyPairTypesResponse
        """
        # ToDo: implement KeyPairResponse-model
        return self.request(
            "keypair/types",
            "GET",
        )

    def getError(self, errorId: str) -> dict:
        """Get information about an error

        :param errorId: The error id (e.g. p-err-abcdefghij1234567rstuvwyxyz)
        """
        if not isinstance(errorId, str):
            raise TypeError("Expected a errorId of type str. Got %r" % type(errorId))
        return self.request(
            "errors/%s" % errorId,
            "GET",
        )

    def createCustomer(self, customer):
        """Creating a customer

        :param customer: Customer object
        :type customer: Customer
        :return: The created customer object
        :rtype: Customer
        """
        if not isinstance(customer, Customer):
            raise TypeError("Expected a Customer object. Got %r" % type(customer))
        if customer.key:
            raise TypeError("Customer has a id (key) set. "
                            "Call updateCustomer to update it or remove it to create a new one.")
        data = self.request(
            "customers",
            "POST",
            customer.serialize(),
        )
        # API docs wrong: we get only a dict with the id back
        return self.getCustomer(data["id"])

    def updateCustomer(self, customer):
        """Update a customer using unique customerId or the resource id from the customers resource.
        The customer MUST have customerId oder key (id)

        :param customer: Customer object
        :type customer: Customer
        :return: The updated customer object
        :rtype: Customer
        """
        if not isinstance(customer, Customer):
            raise TypeError("Expected a Customer object. Got %r" % type(customer))
        if not customer.keyOrCustomerId:
            raise TypeError("Customer has no customerId oder key (id)")
        data = self.request(
            "customers/%s" % customer.keyOrCustomerId,
            "PUT",
            customer.serialize(),
        )
        # API docs wrong: we get only a dict with the id back
        return self.getCustomer(data["id"])

    def createOrUpdateCustomer(self, customer):
        try:
            return self.createCustomer(customer)
        except ErrorResponse as er:
            if er.errors and er.statusCode == 400 and er.errors[0].code == "API.410.200.010":
                return self.updateCustomer(customer)
            raise er

    def deleteCustomer(self, customer):
        """Delete a customer using unique customerId or the resource id from the customers resource.
        The customer MUST have customerId oder key (id)

        :param customer: Customer object, customerId or id (key)
        :type customer: Customer or str
        :return: The id of the customer
        :rtype: str
        """
        if isinstance(customer, Customer):
            if not customer.key and not customer.customerId:
                raise TypeError("Customer has no customerId oder key (id)")
            codeOrExternalId = customer.customerId or customer.key
        elif isinstance(customer, str):
            codeOrExternalId = customer
        else:
            raise TypeError("Expected a Customer object or str. Got %r" % type(customer))
        data = self.request(
            "customers/%s" % codeOrExternalId,
            "DELETE",
        )
        return data["id"]

    def getCustomer(self, codeOrExternalId):
        """Fetch a customer using unique customerId or the resource id from the customers resource.

        :param codeOrExternalId: customerId or id (key)
        :type codeOrExternalId: str
        :return: The fetched customer object
        :rtype: Customer
        """
        data = self.request(
            "customers/%s" % codeOrExternalId,
            "GET",
        )
        return Customer.fromDict(data)

    def createBasket(self, basket):
        """Creating a basket

        :param basket: Basket object
        :type basket: Basket
        :return: The created Basket object
        :rtype: Basket
        """
        if not isinstance(basket, Basket):
            raise TypeError("Expected a Basket object. Got %r" % type(basket))
        data = self.request(
            "baskets",
            "POST",
            basket.serialize(),
        )
        return self.getBasket(data["id"])

    def updateBasket(self, basket):
        """Update a basket.
        The basket MUST have key (id)

        :param basket: Basket object
        :type basket: Basket
        :return: The updated basket object
        :rtype: Basket
        """
        if not isinstance(basket, Basket):
            raise TypeError("Expected a Basket object. Got %r" % type(basket))
        if not basket.key:
            raise TypeError("Basket has no key (id)")
        data = self.request(
            "baskets/%s" % basket.key,
            "PUT",
            basket.serialize(),
        )
        return self.getBasket(data["id"])

    def getBasket(self, basketId):
        """Fetch a basket.

        :param basketId: basket's id (key)
        :type basketId: str
        :return: The fetched basket object
        :rtype: Basket
        """
        data = self.request(
            "baskets/%s" % basketId,
            "GET",
        )
        return Basket.fromDict(data)

    def createPaymentType(self, paymentType):
        """Create a new PaymentType at Unzer.

        This can be any Object which inherits the abstract class PaymentType.

        :param paymentType: The PaymentPage model
        :type paymentType: PaymentType
        :return: The paymentType response
        :rtype: PaymentType
        """
        if not isinstance(paymentType, PaymentType):
            raise TypeError("Expected a PaymentType object. Got %r" % type(paymentType))
        paymentType.validateBeforeRequest()
        data = self.request(
            "types/%s" % paymentType.method,
            "POST",
            paymentType.serialize(),
        )
        return type(paymentType).fromDict(data)

    def createPaymentPage(self, paymentPage):
        """The initialize payment page call with direct charge purpose.

        :param paymentPage: The PaymentPage model
        :type paymentPage: PaymentPage
        :return: The PaymentPageResponse
        :rtype: PaymentPageResponse
        """
        if not isinstance(paymentPage, PaymentPage) or isinstance(paymentPage, PaymentPageResponse):
            raise TypeError("Expected a PaymentPage object. Got %r" % type(paymentPage))
        paymentPage.validateBeforeRequest()
        data = self.request(
            "paypage/%s" % paymentPage.action,
            "POST",
            paymentPage.serialize(),
        )
        return PaymentPageResponse.fromDict(data)

    def getPaymentPage(self, payPageId):
        """Fetch the payment resource. Provides an overview about a payment.

        :param payPageId: The related payment page id.
        :type payPageId: str
        :return: The PaymentPage ressource
        :rtype: PaymentPageResponse
        """
        if not isinstance(payPageId, str):
            raise TypeError("Expected a payPageId of type str. Got %r" % type(payPageId))
        data = self.request(
            "paypage/%s" % payPageId,
            "GET",
        )
        return PaymentPageResponse.fromDict(data)

    def getPayment(self, codeOrOrderId):
        """Fetch the payment resource. Provides an overview about a payment.

        :param codeOrOrderId: The id of the order
        :type codeOrOrderId: str
        :return: Payment ressource
        :rtype: PaymentGetResponse
        """
        if not isinstance(codeOrOrderId, str):
            raise TypeError("Expected a codeOrOrderId of type str. Got %r" % type(codeOrOrderId))
        data = self.request(
            "payments/%s" % codeOrOrderId,
            "GET",
        )
        return PaymentGetResponse.fromDict(data, self)

    def authorize(self, payment) -> PaymentResponse:
        """Authorize call for redirect payments.

        The paymentType will be created within this method,
        if not already created.

        :param payment: The PaymentRequest model
        :type payment: PaymentRequest
        :return: The paymentType response
        :rtype: PaymentResponse
        """
        return self._authorize_or_charge("authorize", payment)

    def charge(self, payment):
        """Charge call for redirect payments.

        The paymentType will be created within this method,
        if not already created.

        :param payment: The PaymentRequest model
        :type payment: PaymentRequest
        :return: The paymentType response
        :rtype: PaymentResponse
        """
        return self._authorize_or_charge("charges", payment)

    def _authorize_or_charge(self, type_: str, payment: PaymentRequest) -> PaymentResponse:
        """Internal helper for authorize and charge calls
        """
        if type_ not in {"authorize", "charges"}:
            raise ValueError("Invalid type %r" % type_)
        if not isinstance(payment, PaymentRequest):
            raise TypeError("Expected a PaymentRequest object. Got %r" % type(PaymentRequest))
        if not payment.paymentType:
            raise ValueError("No paymentType set")
        if not payment.paymentType.key:
            payment.paymentType = self.createPaymentType(payment.paymentType)
        payment.validateBeforeRequest()
        data = self.request(
            "/".join(filter(None, ["payments", payment.paymentId, type_])),
            "POST",
            payment.serialize(),
        )
        if data.get("isError"):
            raise ErrorResponse.fromDict(data)
        return PaymentResponse.fromDict(data, self)

    def getChargedTransaction(self, codeOrOrderId, txnCode):
        """Fetch the corresponding charged transaction.
        The first found charged transaction will be returned if the <txnCode> = null.

        :param codeOrOrderId: The id of the payment
        :type codeOrOrderId: str
        :param txnCode: The id of the transaction
        :type txnCode: str

        :return: PaymentResponse ressource
        :rtype: PaymentResponse
        """
        if not isinstance(codeOrOrderId, str):
            raise TypeError("Expected a codeOrOrderId of type str. Got %r" % type(codeOrOrderId))
        if not isinstance(txnCode, (str, NoneType)):
            raise TypeError("Expected a txnCode of type str or None. Got %r" % type(txnCode))
        data = self.request(
            "payments/%s/charges/%s" % (codeOrOrderId, txnCode or ""),
            "GET",
        )
        return PaymentResponse.fromDict(data, self)

    def listWebhooks(self):
        """Get all webhook resources.

        :return: A list of Webhooks
        :rtype: list[Webhook]
        """
        data = self.request(
            "webhooks",
            "GET",
        )
        return self._loadWebhookResponse(data)

    def getWebhook(self, webhookId):
        """Get one specific webhook resource.

        :param webhookId: The id of the webhook.
        :type webhookId: str
        :return: The webhook resource.
        :rtype: Webhook
        """
        data = self.request(
            "webhooks/%s" % webhookId,
            "GET",
        )
        return Webhook.fromDict(data)

    def createWebhook(self, webhook):
        """Create a new webhook.

        :param webhook: The webhook mode
        :return: A list of created Webhooks models (each for each event-type)
        :rtype: list[Webhook]
        """
        if not isinstance(webhook, Webhook):
            raise TypeError("Expected a Webhook object. Got %r" % type(webhook))
        if webhook.webhookId:
            raise TypeError("Webhook has a id set. "
                            "Call updateWebhook to update it or remove the id to create a new one.")
        webhook.validateBeforeRequest()
        data = self.request(
            "webhooks",
            "POST",
            webhook.serialize(),
        )
        return self._loadWebhookResponse(data)

    def updateWebhook(self, webhook):
        """Update the URL for an existing webhook.
        Will not change the event (not supported by unzer-api)!

        :param webhook: The webhook resource to be updated
        :type webhook: Webhook
        :return: The updated webhook
        :rtype: Webhook
        """
        if not isinstance(webhook, Webhook):
            raise TypeError("Expected a Webhook object. Got %r" % type(webhook))
        if not webhook.webhookId:
            raise ValueError("Webhook to update has no id")
        if not webhook.url:
            raise ValueError("Webhook to update has no url")
        data = self.request(
            "webhooks/%s" % webhook.webhookId,
            "PUT",
            {"url": webhook.url},
        )
        return Webhook.fromDict(data)

    def _loadWebhookResponse(self, data):
        """Helper method load webhook responses.

        :param data: The data from the request.
        :type data: dict
        :return: A list of Webhooks
        :rtype: list[Webhook]
        """
        if "events" not in data:
            webhooks = [data]  # got exactly one webhook, data is the webhook itself
        else:
            webhooks = data["events"]  # list of webhooks wrapped in events property
        return map(Webhook.fromDict, webhooks)

    def deleteWebhook(self, webhookOrId):
        """Delete a specific webhook.

        :param webhookOrId: A webhook id or webhook model
        :type webhookOrId: str | Webhook
        :return: The id of the deleted webhook
        :type: str
        """
        if isinstance(webhookOrId, Webhook):
            webhookOrId = webhookOrId.webhookId
        data = self.request(
            "webhooks/%s" % webhookOrId,
            "DELETE",
        )
        return data["id"]

    def deleteAllWebhooks(self):
        """Delete all webhooks

        :return: A list of the deleted webhooks
        :rtype: list[dict]
        """
        data = self.request(
            "webhooks",
            "DELETE",
        )
        return data["events"]



from unzer.model.payment import PaymentMethodTypes, PaymentTypes
from .abstract_paymenttype import PaymentType


class PaylaterInvoice(PaymentType):
    """Paylater Unzer Invoice

    Unzer Invoice is a part of Unzer's Buy Now Pay Later (BNPL) offering.
    """

    method = PaymentTypes.UNZER_PAYLATER_INVOICE
    method_name = PaymentMethodTypes.UNZER_INVOICE



from .base import BaseModel


class Address(BaseModel):
    def __init__(
            self,
            firstname,
            lastname,
            street=None,
            state=None,
            zipCode=None,
            city=None,
            country=None,
            **kwargs
    ):
        """Create a new Address.

        :param firstname: (optional) Address firstname (+lastname: max. 81 chars). Required in case of billing address.
        :type firstname: str
        :param lastname: (optional)  Address lastname (+firstname: max. 81 chars). Required in case of billing address.
        :type lastname: str
        :param street: (optional) Address street (max. 50 chars). Required in case of billing address.
        :type street: str
        :param state: (optional) Address state in ISO 3166-2 format (max. 8 chars). Required in case of billing address.
        :type state: str
        :param zipCode: (optional) Address zip code (max. 10 chars). Required in case of billing address.
        :type zipCode: str
        :param city: (optional) Address city (max. 30 chars). Required in case of billing address.
        :type city: str
        :param country: (optional) Address country in ISO A2 format (max. 2 chars). Required in case of billing address.
        :type country: str
        """
        super().__init__(**kwargs)
        self.firstname = firstname  # type: str
        self.lastname = lastname  # type: str
        self.street = street  # type: str
        self.state = state  # type: str
        self.zipCode = zipCode  # type: str
        self.city = city  # type: str
        self.country = country  # type: str

    @property
    def name(self):
        return "%s %s" % (
            self.getString(self.firstname),
            self.getString(self.lastname),
        )

    @name.setter
    def name(self, name):
        try:
            self.firstname, self.lastname = name.split(" ", 1)
        except ValueError:
            self.firstname, self.lastname = name, None

    def serialize(self):
        return {
            "name": self.getString(self.name),
            "street": self.getString(self.street),
            "state": self.getString(self.state),
            "zip": self.getString(self.zipCode),
            "city": self.getString(self.city),
            "country": self.getString(self.country),
        }

    @classmethod
    def fromDict(cls, data):
        try:
            firstname, lastname = data["name"].split(" ", 1)
        except ValueError:
            firstname, lastname = data["name"], None
        return cls(
            firstname=firstname,
            lastname=lastname,
            street=data["street"],
            state=data["state"],
            zipCode=data["zip"],
            city=data["city"],
            country=data["country"],
        )



from .base import BaseModel
from .payment import Action
from ..utils import parseBool


class PaymentPage(BaseModel):
    REQUIRED_ATTRIBUTES = [
        "action",
        "amount",
        "currency",
        "returnUrl",
    ]

    def __init__(
            self,
            action=None,
            amount=None,
            currency="EUR",
            invoiceId=None,
            orderId=None,
            card3ds=None,
            returnUrl=None,
            excludeTypes=None,
            additionalAttributes=None,
            logoImage=None,
            fullPageImage=None,
            shopName=None,
            shopDescription=None,
            tagline=None,
            css=None,
            termsAndConditionUrl=None,
            privacyPolicyUrl=None,
            imprintUrl=None,
            helpUrl=None,
            contactUrl=None,
            customerId=None,
            metadataId=None,
            basketId=None,
            **kwargs
    ):
        """Create a new PaymentPage.

        Payment attributes:
        :param action: (required) Action for this paypage. Charge or authorize.
        :type action: str | Action
        :param amount: (required) The transaction amount.
        :type amount: float
        :param currency: (required) The transaction currency, in the ISO 4217 alpha-3 format.
        :type currency: str
        :param invoiceId: (optional) Your internal invoice ID.
        :type invoiceId: str
        :param orderId: (optional) A unique order ID that identifies the payment on your side.
        :type orderId: str
        :param card3ds: (optional) Switches between 3ds and non 3ds card transactions.
        :type card3ds: bool
        :param returnUrl: (required) The URL to redirect the customer to after the payment is completed.
        :type returnUrl: str
        :param excludeTypes: (optional) Exclude some of the payment types from the Payment Page.
        :type excludeTypes: list[str]
        :param additionalAttributes: (optional) Attributes for LinkPay.
        :type additionalAttributes: dict[str, str]

        Paypage config:
        :param logoImage: (optional) Your company logo to show in the Embedded Payment Page’s header.
        :type logoImage: str
        :param fullPageImage: (optional) The URL of the image to show in the Hosted Payment Page’s background.
        :type fullPageImage: str
        :param shopName: (optional) Your company name to show in the Embedded Payment Page’s header.
        :type shopName: str
        :param shopDescription: (optional) Main description of the purchase.
        :type shopDescription: str
        :param tagline: (optional) A short description to show in the Payment Page’s header.
        :type tagline: str
        :param css: (optional)
        :type css: dict[str, str]
        :param termsAndConditionUrl: (optional) Your Terms and Conditions URL to show in the Payment Page’s footer.
        :type termsAndConditionUrl: str
        :param privacyPolicyUrl: (optional) Your Privacy Policy URL to show in the Payment Page’s footer.
        :type privacyPolicyUrl: str
        :param imprintUrl: (optional) Your imprint URL to show in the Payment Page’s footer.
        :type imprintUrl: str
        :param helpUrl: (optional) The URL of the help page to show in the Payment Page’s header.
        :type helpUrl: str
        :param contactUrl: (optional) The URL of the contact page to show in the Payment Page’s header.
        :type contactUrl: str

        Resources
        :param customerId: (optional) The ID of the customers resource to be used.
        :type customerId: str
        :param metadataId: (optional) The ID of the metadata resource to be used.
        :type metadataId: str
        :param basketId: (optional) The ID of the baskets resource to be used.
        :type basketId: str
        """
        super().__init__(**kwargs)
        if excludeTypes is None:
            excludeTypes = []
        if additionalAttributes is None:
            additionalAttributes = {}
        if css is None:
            css = {}
        if action not in {Action.CHARGE, Action.AUTHORIZE}:
            raise TypeError("Invalid action %r" % action)
        if not isinstance(card3ds, bool):
            raise TypeError("Invalid value %r for card3ds. Must be a boolean." % card3ds)
        self.amount = amount  # type:float
        self.currency = currency  # type:str
        self.returnUrl = returnUrl  # type:str
        self.logoImage = logoImage  # type:str
        self.fullPageImage = fullPageImage  # type:str
        self.shopName = shopName  # type:str
        self.shopDescription = shopDescription  # type:str
        self.tagline = tagline  # type:str
        self.css = css  # type:dict[str, str]
        self.termsAndConditionUrl = termsAndConditionUrl  # type:str
        self.privacyPolicyUrl = privacyPolicyUrl  # type:str
        self.imprintUrl = imprintUrl  # type:str
        self.helpUrl = helpUrl  # type:str
        self.contactUrl = contactUrl  # type:str
        self.invoiceId = invoiceId  # type:str
        self.orderId = orderId  # type:str
        self.card3ds = card3ds  # type:bool
        self.additionalAttributes = additionalAttributes  # type:dict[str, str]
        self.excludeTypes = excludeTypes  # type:list[str]
        self.action = action  # type:str
        self.customerId = customerId  # type:str
        self.metadataId = metadataId  # type:str
        self.basketId = basketId  # type:str

    def serialize(self):
        data = {
            "amount": self.amount,
            "currency": self.currency,
            "invoiceId": self.invoiceId,
            "orderId": self.orderId,
            "card3ds": self.card3ds,
            "returnUrl": self.returnUrl,
            "excludeTypes": self.excludeTypes or [],
            "additionalAttributes": self.additionalAttributes,
            "logoImage": self.logoImage or None,
            "fullPageImage": self.fullPageImage or None,
            "shopName": self.shopName,
            "shopDescription": self.shopDescription,
            "tagline": self.tagline,
            "css": self.css,
            "termsAndConditionUrl": self.termsAndConditionUrl,
            "privacyPolicyUrl": self.privacyPolicyUrl,
            "imprintUrl": self.imprintUrl,
            "helpUrl": self.helpUrl,
            "contactUrl": self.contactUrl,
            "resources": {
                "customerId": self.customerId,
                "basketId": self.basketId,
                "metadataId": self.metadataId,
            },
        }
        return data

    @classmethod
    def fromDict(cls, data):
        raise NotImplementedError("Use PaymentPageResponse.fromDict for your responses.")


class PaymentPageResponse(PaymentPage):
    def __init__(
            self,
            payPageId=None,
            paymentId=None,
            redirectUrl=None,
            billingAddressRequired=None,
            shippingAddressRequired=None,
            **kwargs
    ):
        """Create a new PaymentPageResponse.

        To PaymentPage additionally response only params:
        :param payPageId: Id of this payment page
        :type payPageId: str
        :param paymentId: (optional) The ID of the related payment resource. (After success payment)
        :type paymentId: str
        :param redirectUrl: (optional) A unique Hosted Payment Page URL, where the customer completes the payment.
        :type redirectUrl: str
        :param billingAddressRequired: (optional) Determines whether the customer needs to provide a billing address.
        :type billingAddressRequired: bool
        :param shippingAddressRequired: (optional) Determines whether the customer needs to provide a shipping address.
        :type shippingAddressRequired: bool
        """
        super().__init__(**kwargs)
        self.payPageId = payPageId  # type:str
        self.paymentId = paymentId  # type:str
        self.redirectUrl = redirectUrl  # type:str
        self.billingAddressRequired = billingAddressRequired  # type:bool
        self.shippingAddressRequired = shippingAddressRequired  # type:bool

    def serialize(self):
        data = super().serialize()
        data["id"] = self.payPageId
        data["paymentId"] = self.paymentId
        data["redirectUrl"] = self.redirectUrl
        data["billingAddressRequired"] = self.billingAddressRequired
        data["shippingAddressRequired"] = self.shippingAddressRequired
        return data

    @classmethod
    def fromDict(cls, data):
        data = data.copy()
        data["payPageId"] = data["id"]
        data["customerId"] = data["resources"].get("customerId")
        data["basketId"] = data["resources"].get("basketId")
        data["metadataId"] = data["resources"].get("metadataId")
        data["paymentId"] = data["resources"].get("paymentId")
        data["card3ds"] = parseBool(data["card3ds"])
        data["shippingAddressRequired"] = parseBool(data["shippingAddressRequired"])
        data["billingAddressRequired"] = parseBool(data["billingAddressRequired"])
        data["action"] = data["action"].lower()  # must be equivalent to enum *Action*
        return cls(**data)



import datetime


def parseBool(value):
    return str(value).lower() == "true"


def parseDateTime(value):
    if not value:
        return None
    if isinstance(value, datetime.datetime):
        return value
    if "-" in value:  # ISO Date
        return datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
    elif "." in value:  # European Date
        return datetime.datetime.strptime(value, "%d.%m.%Y %H:%M:%S")
    raise TypeError("Invalid date format of %r" % value)



from .address import Address
from .basket import Basket
from .basketItem import BasketItem
from .customer import Customer
from .error import Error, ErrorResponse
from .payment import (
    Action,
    PaymentGetResponse,
    PaymentMethodTypes,
    PaymentRequest,
    PaymentResponse,
    PaymentState,
    PaymentTransaction,
    PaymentTypes,
    TransactionStatus,
)
from .payment_type import *
from .paymentpage import PaymentPage, PaymentPageResponse
from .webhook import Events, Webhook

__all__ = [
    "Address",
    "Basket",
    "BasketItem",
    "Customer",
    # error
    "Error",
    "ErrorResponse",
    # payment
    "Action",
    "PaymentGetResponse",
    "PaymentMethodTypes",
    "PaymentRequest",
    "PaymentResponse",
    "PaymentState",
    "PaymentTransaction",
    "PaymentTypes",
    "TransactionStatus",
    # payment_type
    "PaymentType",
    "Applepay",
    "Bancontact",
    "Card",
    "Googlepay",
    "Ideal",
    "PaylaterInvoice",
    "PayPal",
    "Sofort",
    # paymentpage
    "PaymentPage",
    "PaymentPageResponse",
    # webhook
    "Webhook",
    "Events",
]



import datetime

from .address import Address
from .base import BaseModel


class Salutation:
    MR = "mr"
    MRS = "mrs"
    UNKNOWN = "unknown"


class Customer(BaseModel):

    def __init__(
            self,
            firstname,
            lastname,
            salutation=None,
            key=None,
            customerId=None,
            birthDate=None,
            email=None,
            phone=None,
            mobile=None,
            billingAddress=None,
            shippingAddress=None,
            company=None,
            companyData=None,
            **kwargs
    ):
        """Create a new Customer.

        :param key: (optional) (original: id) Customer's generated code by Unzer's Payment
        :type key: str
        :param firstname: Customer's lastname
        :type firstname: str
        :param lastname: Customer's firstname
        :type lastname: str
        :param salutation: (optional) Must be either 'mr', 'mrs' or 'unknown'
        :type salutation: str | Salutation
        :param company: (optional) Company name
        :type company: str
        :param customerId: (optional) Must be unique and identifies the customer.
            Can be used in place of the resource id
        :type customerId: str
        :param birthDate: (optional) Birthdate of the customer in format yyyy-mm-dd or dd.mm.yyyy
        :type birthDate: datetime.datetime | datetime.date | str
        :param email: (optional) Customer's email
        :type email: str
        :param phone: (optional) Customer's phone
        :type phone: str
        :param mobile: (optional) Customer's mobile
        :type mobile: str
        :param billingAddress: (optional) billing address
        :type billingAddress: Address
        :param shippingAddress: (optional) shipping address
        :type shippingAddress: Address
        :param companyData: (optional)
        :type companyData: CompanyInfo
        """
        super().__init__(**kwargs)
        if salutation is None:
            salutation = Salutation.UNKNOWN
        elif salutation not in {Salutation.MR, Salutation.MRS, Salutation.UNKNOWN}:
            raise TypeError("Invalid salutation")
        self.key = key  # type: str
        self.firstname = firstname  # type: str
        self.lastname = lastname  # type: str
        self.salutation = salutation  # type: Salutation
        self.customerId = customerId  # type: str
        self.birthDate = birthDate  # type: datetime.datetime
        self.email = email  # type: str
        self.phone = phone  # type: str
        self.mobile = mobile  # type: str
        self.billingAddress = billingAddress  # type: Address
        self.shippingAddress = shippingAddress  # type: Address
        self.company = company  # type: str
        self.companyData = companyData  # type: CompanyInfo

    @property
    def keyOrCustomerId(self):
        return self.key or self.customerId

    @property
    def salutation(self):
        return self._salutation

    @salutation.setter
    def salutation(self, value):
        if not value:
            value = Salutation.UNKNOWN
        elif value not in {Salutation.MR, Salutation.MRS, Salutation.UNKNOWN}:
            raise TypeError("Invalid salutation %r" % value)
        self._salutation = value

    @property
    def birthDate(self):
        return self._birthDate

    @birthDate.setter
    def birthDate(self, value):
        if not value:
            value = None
        elif isinstance(value, str):
            if "-" in value:  # ISO Date
                value = datetime.datetime.strptime(value, "%Y-%m-%d")
            elif "." in value:  # European Date
                value = datetime.datetime.strptime(value, "%d.%m.%Y")
            else:
                raise TypeError("Invalid date format of %r" % value)
        elif not isinstance(value, (datetime.datetime, datetime.date)):
            raise TypeError("Invalid value %r" % value)
        self._birthDate = value

    @property
    def phone(self):
        return self._phone

    @phone.setter
    def phone(self, value):
        if not value:
            value = None
        self._phone = value

    @property
    def mobile(self):
        return self._mobile

    @mobile.setter
    def mobile(self, value):
        if not value:
            value = None
        self._mobile = value

    def serialize(self):
        birthDate = self.birthDate
        if isinstance(birthDate, (datetime.datetime, datetime.date)):
            birthDate = birthDate.strftime("%Y-%m-%d")

        billingAddress = self.billingAddress
        if billingAddress is None:
            billingAddress = self.EMPTY_STRING
        else:
            assert isinstance(billingAddress, Address)
            billingAddress = billingAddress.serialize()

        shippingAddress = self.shippingAddress
        if shippingAddress is None:
            shippingAddress = self.EMPTY_STRING
        else:
            assert isinstance(shippingAddress, Address)
            shippingAddress = shippingAddress.serialize()

        return {
            "lastname": self.lastname,
            "firstname": self.firstname,
            "id": self.getString(self.key),
            "salutation": self.getString(self.salutation),
            "company": self.getString(self.company),
            "customerId": self.getString(self.customerId),
            "birthDate": self.getString(birthDate),
            "email": self.getString(self.email),
            "phone": self.getString(self.phone),
            "mobile": self.getString(self.mobile),
            "billingAddress": billingAddress,
            "shippingAddress": shippingAddress,

            # Additional information for B2B Customer #ToDo
            # "companyInfo": {
            # 	# Mandatory in case companyInfo is existing, restrict '<' and '>'
            # 	"registrationType": "registered|not_registered",
            # 	# Mandatory for REGISTERED, restrict '<' and '>'
            # 	"commercialRegisterNumber": "...",
            # 	# Mandatory must be the value "OWNER" for NOT_REGISTERED, restrict '<' and '>'
            # 	"function": "...",
            # 	# Mandatory for NOT_REGISTERED, restrict '<' and '>'
            # 	"commercialSector": "..."
            # }
        }

    @classmethod
    def fromDict(cls, data):
        data = data.copy()
        data["key"] = data["id"]
        data["billingAddress"] = Address.fromDict(data["billingAddress"])
        data["shippingAddress"] = Address.fromDict(data["shippingAddress"])
        return cls(**data)



import datetime
import logging

import requests

logger = logging.getLogger("unzer-sdk")


class Error:
    def __init__(self, code, merchantMessage, customerMessage, **kwargs):
        self.code = code
        self.merchantMessage = merchantMessage
        self.customerMessage = customerMessage
        if kwargs:
            logger.warning("Error got additional unhandled data: %r", kwargs)

    def __str__(self):
        return "%s %s: %s" % (self.__class__.__name__, self.code, self.merchantMessage)

    def __repr__(self):
        return "%s.%s(code=%r, merchantMessage=%r, customerMessage=%r)" % (
            self.__class__.__module__,
            self.__class__.__name__,
            self.code,
            self.merchantMessage,
            self.customerMessage,
        )


class ErrorResponse(Exception):
    def __init__(
            self,
            message,
            timestamp=None,
            url=None,
            errors=None,
            errorId=None,
            statusCode=0,
            traceId=None,
            isError=None,
            isPending=None,
            isSuccess=None,
            srcResponse=None,
            **kwargs

    ):
        super(ErrorResponse, self).__init__(message)
        if errors is None:
            errors = []
        self.timestamp = timestamp  # type: str
        self.url = url  # type: str
        self.errors = errors  # type: list[Error]
        self.errorId = errorId  # type: str
        self.statusCode = statusCode  # type: int
        self.traceId = traceId  # type: str
        self.isError = isError  # type: bool
        self.isPending = isPending  # type: bool
        self.isSuccess = isSuccess  # type: bool
        self.srcResponse = srcResponse  # type: requests.Response
        if kwargs:
            logger.warning("ErrorResponse got additional unhandled data: %r", kwargs)

    @classmethod
    def fromDict(cls, data, message="Unzer Error"):
        return cls(
            message,
            timestamp=datetime.datetime.strptime(data["timestamp"], "%Y-%m-%d %H:%M:%S"),
            url=data["url"],
            errors=[Error(**error) for error in data["errors"]],
            errorId=data.get("id"),
            traceId=data.get("traceId"),
            isError=data.get("isError"),
            isPending=data.get("isPending"),
            isSuccess=data.get("isSuccess"),
        )

    def __repr__(self):
        return "%s.%s(url=%r, errorId=%r, traceId=%r, errors=%r)" % (
            self.__class__.__module__,
            self.__class__.__name__,
            self.url,
            self.errorId,
            self.traceId,
            self.errors,
        )



from unzer.model.payment import PaymentMethodTypes, PaymentTypes
from .abstract_paymenttype import PaymentType


class Googlepay(PaymentType):
    method = PaymentTypes.GOOGLE_PAY
    method_name = PaymentMethodTypes.GOOGLE_PAY



from .abstract_paymenttype import PaymentType
from .applepay import Applepay
from .bancontact import Bancontact
from .card import Card
from .googlepay import Googlepay
from .ideal import Ideal
from .paylater_invoice import PaylaterInvoice
from .paypal import PayPal
from .sofort import Sofort

__all__ = [
    "PaymentType",
    "Applepay",
    "Bancontact",
    "Card",
    "Googlepay",
    "Ideal",
    "PaylaterInvoice",
    "PayPal",
    "Sofort",
]



from unzer.model.payment import PaymentMethodTypes, PaymentTypes
from .abstract_paymenttype import PaymentType


class Card(PaymentType):
    method = PaymentTypes.CARD
    method_name = PaymentMethodTypes.CARD



import typing as t

from unzer.model.payment import PaymentMethodTypes, PaymentTypes
from .abstract_paymenttype import PaymentType


class Bancontact(PaymentType):
    method = PaymentTypes.BANCONTACT
    method_name = PaymentMethodTypes.BANCONTACT

    REQUIRED_ATTRIBUTES = ["holder"]

    def __init__(
            self,
            holder: str | None = None,
            **kwargs
    ):
        """Create a new Bancontact paymentType ressource.

        :param holder: The holder name.
        """
        super().__init__(**kwargs)
        self.holder: str | None = holder

    def serialize(self) -> dict:
        if not self.holder:
            return super().serialize()

        return {
            "holder": self.holder,
        }

    @classmethod
    def fromDict(cls, data: dict) -> t.Self:
        data = data.copy()
        data["key"] = data["id"]
        return cls(**data)



from .base import BaseModel

# https://docs.unzer.com/reference/basic-integration-req/#allowlist-of-ip-addresses
IP_ADDRESS = {
    # Production environment
    "3.120.62.83", "18.196.35.212",
    # Sandbox environment
    "35.157.10.171", "18.197.240.190",
}


class Events:
    ALL = "all"
    AUTHORIZE = "authorize"
    AUTHORIZE_SUCCEEDED = "authorize.succeeded"
    AUTHORIZE_FAILED = "authorize.failed"
    AUTHORIZE_PENDING = "authorize.pendin"
    AUTHORIZE_EXPIRED = "authorize.expired"
    AUTHORIZE_CANCELED = "authorize.canceled"
    CHARGE = "charge"
    CHARGE_SUCCEEDED = "charge.succeeded"
    CHARGE_FAILED = "charge.failed"
    CHARGE_PENDING = "charge.pendin"
    CHARGE_EXPIRED = "charge.expired"
    CHARGE_CANCELED = "charge.canceled"
    CHARGEBACK = "chargeback"
    TYPES = "types"
    CUSTOMER = "customer"
    CUSTOMER_CREATED = "customer.created"
    CUSTOMER_DELETED = "customer.deleted"
    CUSTOMER_UPDATED = "customer.updated"
    BASKET = "basket"
    BASKET_CREATE = "basket.created"
    BASKET_UPDATED = "basket.updated"
    BASKET_USED = "basket.used"
    PAYMENT = "payment"
    PAYMENT_PENDING = "payment.pending"
    PAYMENT_COMPLETED = "payment.completed"
    PAYMENT_CANCELED = "payment.canceled"
    PAYMENT_PARTLY = "payment.partly"
    PAYMENT_REVIEW = "payment.payment_review"
    PAYMENT_CHARGEBACK = "payment.chargeback"
    SHIPMENT = "shipment"
    PAYOUT = "payout"
    PAYOUT_SUCCEEDED = "payout.succeeded"
    PAYOUT_FAILED = "payout.failed"


class Webhook(BaseModel):
    REQUIRED_ATTRIBUTES = [
        "event",
        "url"
    ]

    def __init__(
            self,
            url,
            event=Events.ALL,
            webhookId=None,
            **kwargs
    ):
        """Create a new Webhook.

        :param webhookId: The id of the webhook
        :type webhookId: str
        :param event: The or a list of events for this webhook
        :type event: str
        :param url: The url of the webhook
        :type url: str
        """
        super().__init__(**kwargs)
        self.webhookId = webhookId
        self.event = event
        self.url = url

    @property
    def event(self):
        return self._event

    @event.setter
    def event(self, value):
        if isinstance(value, str):
            value = [value]
        if not isinstance(value, list):
            raise TypeError("Event must be a str or list of str. Got %r" % type(value))
        for val in value:
            if val not in vars(Events).values():
                raise TypeError("Invalid value %r for event" % value)
        self._event = value

    def serialize(self):
        return {
            "eventList": self.event,
            "url": self.url,
        }

    @classmethod
    def fromDict(cls, data):
        data = data.copy()
        data["webhookId"] = data["id"]
        return cls(**data)



import abc

import typing as t

if t.TYPE_CHECKING:
    from ..client import UnzerClient  # noqa # pylint: disable=unused-import


class BaseModel(abc.ABC):
    EMPTY_STRING = ""

    REQUIRED_ATTRIBUTES = []

    def __init__(
            self,
            client: "UnzerClient" = None,
            **kwargs,
    ):
        """
        :param client: (optional) The client instance.
        """
        super().__init__()
        self._client: "UnzerClient" = client

    def getString(self, value):
        if value is None:
            return self.EMPTY_STRING
        return value

    @abc.abstractmethod
    def serialize(self):
        """Serialize data from an object as dict for the request-payload."""
        pass

    @classmethod
    @abc.abstractmethod
    def fromDict(cls, data):
        """Unserialize data from a dict from a response to new object"""
        pass

    def validateBeforeRequest(self):
        """Validate the model.

        Useful to check the model for validity before the API request.
        By default, check for the required attributes,
        set in :attr:`REQUIRED_ATTRIBUTES` (class attribute).
        """
        for attr in type(self).REQUIRED_ATTRIBUTES:  # use always the cls-attributes
            if not getattr(self, attr):
                raise ValueError("%s misses the attribute *%s*." % (type(self).__name__, attr))
        return True

    def __repr__(self):
        return "%s.%s(%s)" % (
            self.__class__.__module__,
            self.__class__.__name__,
            ", ".join("%s=%r" % (k, v) for k, v in sorted(self))
        )

    def asDict(self):
        """Return the model as dict.

        This will not be done recursive.
        """
        # instance attributes
        data = {k: v for k, v in vars(self).items() if not k.startswith("_")}
        # class properties
        for k, v in vars(self.__class__).items():
            if isinstance(v, property):
                data[k] = getattr(self, k)
        return data

    def __iter__(self):
        """Yield the attributes of the model"""
        for k, v in self.asDict().items():
            yield k, v



from unzer.model.payment import PaymentMethodTypes, PaymentTypes
from .abstract_paymenttype import PaymentType


class Ideal(PaymentType):
    method = PaymentTypes.IDEAL
    method_name = PaymentMethodTypes.IDEAL



from .base import BaseModel
from .basketItem import BasketItem


class Basket(BaseModel):
    def __init__(
            self,
            key=None,
            amountTotalGross=None,
            amountTotalVat=None,
            amountTotalDiscount=None,
            currencyCode=None,
            orderId=None,
            note=None,
            basketItems=None,
            **kwargs
    ):
        """Create a new Basket.

        :param key: (optional)
        :type key: str
        :param amountTotalGross: (optional)
        :type amountTotalGross: float
        :param amountTotalVat: (optional)
        :type amountTotalVat: float
        :param amountTotalDiscount: (optional)
        :type amountTotalDiscount: float
        :param currencyCode: (optional) example: EUR
        :type currencyCode: str
        :param orderId: example: s-bsk-XXX
        :type orderId: str
        :param note: (optional)
        :type note: str
        :param basketItems: (optional)
        :type basketItems: list[BasketItem]
        """
        super().__init__(**kwargs)
        if basketItems is None:
            basketItems = []
        self.key = key  # type:str
        self.amountTotalGross = amountTotalGross  # type:float
        self.amountTotalVat = amountTotalVat  # type:float
        self.amountTotalDiscount = amountTotalDiscount  # type:float
        self.currencyCode = currencyCode  # type:str
        self.orderId = orderId  # type:str
        self.note = note  # type:str
        self.basketItems = basketItems  # type:list[BasketItem]

    def serialize(self):
        return {
            "id": self.key,
            "amountTotalGross": self.amountTotalGross,
            "amountTotalVat": self.amountTotalVat,
            "amountTotalDiscount": self.amountTotalDiscount,
            "currencyCode": self.getString(self.currencyCode),
            "orderId": self.getString(self.orderId),
            "note": self.getString(self.note),
            "basketItems": [bi.serialize() for bi in self.basketItems],
        }

    @classmethod
    def fromDict(cls, data):
        data = data.copy()
        data["key"] = data["id"]
        data["basketItems"] = [BasketItem.fromDict(basketItem) for basketItem in data["basketItems"]]
        data["amountTotalGross"] = float(data["amountTotalGross"])
        data["amountTotalVat"] = float(data["amountTotalVat"])
        data["amountTotalDiscount"] = float(data["amountTotalDiscount"])
        return cls(**data)



from unzer.model.payment import PaymentMethodTypes, PaymentTypes
from .abstract_paymenttype import PaymentType


class Applepay(PaymentType):
    method = PaymentTypes.APPLE_PAY
    method_name = PaymentMethodTypes.APPLE_PAY



import datetime
import enum
import logging
import re
import typing as t
from types import NoneType

from .base import BaseModel
from ..utils import parseBool, parseDateTime

if t.TYPE_CHECKING:
    from ..client import UnzerClient


class TransactionStatus(enum.Enum):
    SUCCESS = "success"
    PENDING = "pending"
    ERROR = "error"


class Action(enum.Enum):
    CHARGE = "charge"
    AUTHORIZE = "authorize"


class PaymentState(enum.Enum):
    PENDING = 0
    COMPLETED = 1
    CANCELED = 2
    PARTLY = 3
    PAYMENT_REVIEW = 4
    CHARGEBACK = 5
    CREATE = 6


class PaymentTypes(enum.Enum):
    """
    Supported payment types

    Used as short-name in type-ids like ``s-crd-abc456def789``

    source: https://github.com/unzerdev/java-sdk/blob/main/src/main/java/com/unzer/payment/paymenttypes/PaymentTypeEnum.java  # noqa: E501
    """
    CARD = "crd"
    CLICK_TO_PAY = "ctp"
    EPS = "eps"
    GIROPAY = "gro"
    GOOGLE_PAY = "gop"
    IDEAL = "idl"
    INVOICE = "ivc"
    INVOICE_GUARANTEED = "ivg"  # deprecated
    INVOICE_FACTORING = "ivf"  # deprecated
    INVOICE_SECURED = "ivs"  # deprecated
    PAYPAL = "ppl"
    PREPAYMENT = "ppy"
    PRZELEWY24 = "p24"
    SEPA_DIRECT_DEBIT = "sdd"
    SEPA_DIRECT_DEBIT_GUARANTEED = "ddg"  # deprecated
    SEPA_DIRECT_DEBIT_SECURED = "dds"  # deprecated
    SOFORT = "sft"
    PIS = "pis"
    ALIPAY = "ali"
    WECHATPAY = "wcp"
    APPLE_PAY = "apl"
    HIRE_PURCHASE_RATE_PLAN = "hdd"
    INSTALLMENT_SECURED_RATE_PLAN = "ins"  # deprecated
    BANCONTACT = "bct"
    PF_CARD = "pfc"
    PF_EFINANCE = "pfe"
    UNZER_PAYLATER_INVOICE = "piv"
    KLARNA = "kla"
    PAYLATER_INSTALLMENT = "pit"
    PAYLATER_DIRECT_DEBIT = "pdd"
    TWINT = "twt"
    OPEN_BANKING = "obp"
    UNKNOWN = "unknown"


class PaymentMethodTypes(enum.Enum):
    """
    Full name of supported payment types

    Used as name in URLs like ``types/<name>/``

    source: https://github.com/unzerdev/integration-core/blob/master/src/BusinessLogic/Domain/PaymentMethod/Enums/PaymentMethodTypes.php  # noqa: E501
    """
    ALI_PAY = "alipay"
    APPLE_PAY = "applepay"
    BANCONTACT = "bancontact"
    CARD = "card"
    GIROPAY = "giropay"
    GOOGLE_PAY = "googlepay"
    IDEAL = "ideal"
    KLARNA = "klarna"
    PAYPAL = "paypal"
    PAYU = "payu"
    PRZELEWY24 = "przelewy24"
    POST_FINANCE_CARD = "post-finance-card"
    POST_FINANCE_EFINANCE = "post-finance-efinance"
    SOFORT = "sofort"
    TWINT = "twint"
    UNZER_DIRECT_DEBIT = "sepa-direct-debit"
    DIRECT_DEBIT_SECURED = "paylater-direct-debit"
    UNZER_INSTALLMENT = "paylater-installment"
    UNZER_INVOICE = "paylater-invoice"
    UNZER_PREPAYMENT = "prepayment"
    WECHATPAY = "wechatpay"
    EPS = "EPS"
    DIRECT_BANK_TRANSFER = "openbanking-pis"
    CLICK_TO_PAY = "clicktopay"


# TODO: Combine PaymentMethodTypes and PaymentTypes in a dataclass to have their mapping too?

paymentUrlRe = re.compile(
    r"^https://api.unzer.com/v1/"
    r"(?P<operation>[a-z]+)/(?P<paymentId>[\w-]+)"  # payments/{codeOrOrderId}
    r"((/(?P<subOperation>[a-z]+)/(?P<subCode>[\w-]+))?"  # /[charges|authorize|shipments|payouts]/{txnCode|chargeCode}
    r"(/(?P<subSubOperation>[a-z]+)/(?P<subSubCode>[\w-]+))?)?"  # /chargebacks/{chargeBackCode} | /cancels/{cancelCode}
)


class PaymentGetResponse(BaseModel):

    def __init__(
            self,
            paymentId=None,
            paymentType=None,
            state=None,
            currency=None,
            orderId=None,
            invoiceId=None,
            transactions=None,
            card3ds=None,
            amountTotal=None,
            amountCharged=None,
            amountCanceled=None,
            amountRemaining=None,
            customerId=None,
            basketId=None,
            metadataId=None,
            payPageId=None,
            linkPayId=None,
            typeId=None,
            **kwargs
    ):
        """Create a new PaymentGetResponse.

        :param paymentId: The id of payment (ex: s-pay-1), assigned by unzer.
        :type paymentId: str
        :param paymentType: (optional) The type of payment
        :type paymentType: PaymentTypes
        :param state: (optional) Current state of this payment
        :type state: PaymentState
        :param currency: (optional) ISO currency code
        :type currency: str
        :param orderId: (optional) Order id of the merchant application.
            This id can also be used to get payments from the api.
            The id has to be unique for the used key pair.
        :type orderId: str
        :param invoiceId: (optional) InvoiceId of the merchant.
        :type invoiceId: str
        :param transactions: (optional) List of subsequence transaction(s).
        :type transactions: list[PaymentTransaction]
        :param card3ds: (optional)
        :type card3ds: bool | None

        Amounts
        :param amountTotal: (optional) Initial amount reduced by cancellations during authorization
        :type amountTotal: float
        :param amountCharged: (optional) Already charged amount
        :type amountCharged: float
        :param amountCanceled: (optional) Refunded amount of all charges
        :type amountCanceled: float
        :param amountRemaining: (optional) Difference between total and charged
        :type amountRemaining: float

        Resources
        :param customerId: (optional) Customer id used for this transaction.
        :type customerId: str
        :param basketId: (optional) Basket ID used for this transaction.
        :type basketId: str
        :param metadataId: (optional) Meta data ID used for this transaction.
        :type metadataId: str
        :param payPageId: (optional) Payment Page Id related to this payment.
        :type payPageId: str
        :param linkPayId: (optional)
        :type linkPayId: str
        :param typeId: (optional) Id of the types Resource that is to be used for this transaction.
        :type typeId: str
        """
        super().__init__(**kwargs)
        if transactions is None:
            transactions = []
        state = PaymentState(state)
        # if state not in vars(PaymentState).values():
        #     raise TypeError("Invalid state %r" % state)
        if not isinstance(card3ds, (bool, NoneType)):
            raise TypeError("Invalid value %r for card3ds. Must be a boolean or None." % card3ds)
        self.paymentId = paymentId  # type:str
        self.paymentType = paymentType  # type:PaymentTypes
        self.state = state  # type:PaymentState
        self.currency = currency  # type: str
        self.orderId = orderId  # type: str
        self.invoiceId = invoiceId  # type: str
        self.transactions = transactions  # type: list[PaymentTransaction]
        self.card3ds = card3ds  # type: Union[bool, None]
        # Amounts
        self.amountTotal = amountTotal  # type:float
        self.amountCharged = amountCharged  # type:float
        self.amountCanceled = amountCanceled  # type:float
        self.amountRemaining = amountRemaining  # type: float
        # PaymentResponseResources
        self.customerId = customerId  # type: str
        self.paymentId = paymentId  # type: str
        self.basketId = basketId  # type: str
        self.metadataId = metadataId  # type: str
        self.payPageId = payPageId  # type: str
        self.linkPayId = linkPayId  # type: str
        self.typeId = typeId  # type: str

    def serialize(self):
        raise NotImplementedError("No serialisation for response models.")

    # noinspection PyMethodOverriding
    @classmethod
    def fromDict(cls, data, client):
        data = data.copy()
        data["paymentId"] = data["id"]
        if data["resources"].get("typeId"):
            data["paymentType"] = PaymentGetResponse.getPaymentTypeFromTypeId(data["resources"]["typeId"])
        data["state"] = int(data["state"]["id"])
        data["card3ds"] = parseBool(data["card3ds"]) if "card3ds" in data else None
        data["transactions"] = list(map(PaymentTransaction.fromDict, data["transactions"]))
        # Amounts
        data["amountTotal"] = float(data["amount"].get("total", 0))
        data["amountCharged"] = float(data["amount"].get("charged", 0))
        data["amountCanceled"] = float(data["amount"].get("canceled", 0))
        data["amountRemaining"] = float(data["amount"].get("remaining", 0))
        # Resources
        data["customerId"] = data["resources"].get("customerId") or None
        # resources.paymentId is already on top-level
        data["basketId"] = data["resources"].get("basketId") or None
        data["metadataId"] = data["resources"].get("metadataId") or None
        data["payPageId"] = data["resources"].get("payPageId") or None
        data["linkPayId"] = data["resources"].get("linkPayId") or None
        data["typeId"] = data["resources"].get("typeId") or None
        return cls(client=client, **data)

    def getChargedTransactions(self):
        """Fetch the charged transaction of this payment.

        :return:  List of charged transaction resources.
        :rtype: list[PaymentResponse]
        """
        transactions = []
        for txn in filter(lambda txn_: txn_.action == Action.CHARGE, self.transactions):
            transactions.append(self._client.getChargedTransaction(self.paymentId, txn.transactionId))
        return transactions

    @staticmethod
    def getPaymentTypeFromTypeId(typeId) -> PaymentTypes:
        if not typeId:
            raise ValueError("Invalid typeId %r" % typeId)  # TODO: or return PaymentTypes.UNKNOWN?
        paymentType = typeId.split("-")[1].lower()
        try:
            paymentType = PaymentTypes(paymentType)
        except ValueError:
            raise ValueError("Invalid type %r" % typeId)  # TODO: or return PaymentTypes.UNKNOWN?
        return paymentType

    def charge(self, amount: float) -> "PaymentResponse":
        req_kwargs = self.__dict__.copy()
        req_kwargs["paymentType"] = PaymentType.construct(self.paymentType)(self.typeId)
        req_kwargs["amount"] = amount
        logging.debug("req_kwargs = %r", req_kwargs)
        req = PaymentRequest(**req_kwargs)
        return self._client.charge(req)


class PaymentTransaction(BaseModel):
    def __init__(
            self,
            paymentId=None,
            transactionId=None,
            participantId=None,
            date=None,
            action=None,
            status=None,
            url=None,
            amount=None,
            **kwargs
    ):
        """Create a new PaymentGetResponseTransaction.
        :param paymentId: Id of the payment where this transaction belongs to
        :type paymentId: str
        :param transactionId: Id of this transaction (context based to payment)
        :type transactionId: str
        :param participantId: (optional)
        :type participantId: str
        :param date: (optional)
        :type date: datetime.datetime
        :param String:
        :param action: (optional)
        :type action: Action
        :param status: (optional)
        :type status: TransactionStatus
        :param url: (optional)
        :type url: str
        :param amount: (optional)
        :type amount: float
        """
        super().__init__(**kwargs)
        self.paymentId = paymentId  # type:str
        self.transactionId = transactionId  # type:str
        self.participantId = participantId  # type:str
        self.date = date  # type:datetime.datetime
        self.action = action  # type:Action
        self.status = status  # type:TransactionStatus
        self.url = url  # type:str
        self.amount = amount  # type:float

    def serialize(self):
        raise NotImplementedError("No serialisation for response models.")

    @classmethod
    def fromDict(cls, data):
        data = data.copy()
        data["status"] = data["status"].lower()  # must be equivalent to enum *TransactionStatus*
        data["action"] = data["type"].lower()  # must be equivalent to enum *Action*
        data["date"] = parseDateTime(data["date"])
        data["amount"] = float(data["amount"])
        # And now some ugly parsing of the url, because Unzer provide no suitable parameters
        # url-example: https://api.unzer.com/v1/payments/s-pay-123456/charges/s-chg-1
        # url-example: https://api.unzer.com/v1/payments/s-pay-123456/charges/s-chg-1/cancels/s-cnl-1
        assert data["url"], data["url"]
        match = re.match(paymentUrlRe, data["url"])
        assert match, "Cannot match %r" % data["url"]
        matchDict = match.groupdict()
        logging.debug("matchDict: %r for url %r", matchDict, data["url"])
        assert matchDict["operation"] == "payments", "Operation %r not matching" % matchDict["operation"]
        data["paymentId"] = matchDict["paymentId"]
        data["subOperation"] = matchDict["subOperation"]
        data["subCode"] = data["transactionId"] = matchDict["subCode"]
        data["subSubOperation"] = matchDict["subSubOperation"]
        data["subSubCode"] = matchDict["subSubCode"]
        return cls(**data)


class PaymentRequest(BaseModel):
    REQUIRED_ATTRIBUTES = ["paymentType"]

    def __init__(
            self,
            paymentType=None,
            paymentId=None,
            amount=None,
            currency="EUR",
            returnUrl=None,
            card3ds=None,
            paymentReference=None,
            orderId=None,
            invoiceId=None,
            effectiveInterestRate=None,
            customerId=None,
            metadataId=None,
            basketId=None,

            **kwargs
    ):
        """Create a new PaymentRequest.

        :param paymentType: The PaymentType model, will provide the typeId.
        :type paymentType: PaymentType
        :param amount: The amount to be charged on the specified paymentType.
            Amount in positive decimal values. Accepted length: Decimal{10,4}.
        :type amount: float
        :param currency: (optional) ISO currency code.
        :type currency: str
        :param returnUrl: (optional) URL to redirect the customer after
            the payment is completed (in case of redirect payments
            e.g. Paypal, Sofort). Required in condition.
        :type returnUrl: str
        :param card3ds: (optional) Indicate a 3ds transaction.
            Only valid for Card method: Overrides the existing
            credit card configuration if possible.
        :type card3ds: bool
        :param paymentReference: Transaction description
        :type paymentReference: str
        :param orderId: (optional) Order id that identifies the payment on merchant side.
        :type orderId: str
        :param invoiceId: (optional) invoice id that is assigned to the payment on merchant side.
        :type invoiceId: str
        :param effectiveInterestRate: (optional) Only valid for Installment method:
            The affected installment rated. Required in case of Installment method.
        :type effectiveInterestRate: str

        Resources
        :param customerId: (optional) Customer id used for this transaction.
        :type customerId: str
        :param metadataId: (optional) Meta data ID used for this transaction.
        :type metadataId: str
        :param basketId: (optional) Basket ID used for this transaction.
        :type basketId: str
        """
        super().__init__(**kwargs)
        if not isinstance(card3ds, (bool, NoneType)):
            raise TypeError("Invalid value %r for card3ds. Must be a boolean or None." % card3ds)
        self.paymentType = paymentType  # type:PaymentType
        self.paymentId = paymentId  # type:str
        self.amount = amount  # type:float
        self.currency = currency  # type: str
        self.returnUrl = returnUrl  # type: str
        self.card3ds = card3ds  # type: Union[bool, None]
        self.paymentReference = paymentReference  # type: str
        self.orderId = orderId  # type: str
        self.invoiceId = invoiceId  # type: str
        self.effectiveInterestRate = effectiveInterestRate  # type: str
        # PaymentResponseResources
        self.customerId = customerId  # type: str
        self.metadataId = metadataId  # type: str
        self.basketId = basketId  # type: str

    def serialize(self):
        data = {
            "amount": self.amount,
            "currency": self.currency,
            "returnUrl": self.returnUrl,
            "card3ds": self.card3ds,
            "paymentReference": self.paymentReference,
            "orderId": self.orderId,
            "invoiceId": self.invoiceId,
            "effectiveInterestRate": self.effectiveInterestRate,
            "resources": {
                "customerId": self.customerId,
                "typeId": self.paymentType.key if self.paymentType else None,
                "metadataId": self.metadataId,
                "basketId": self.basketId,
            },
        }
        return data

    @classmethod
    def fromDict(cls, data):
        raise NotImplementedError("Use PaymentResponse.fromDict for your responses.")


class PaymentResponse(BaseModel):
    def __init__(
            self,
            transactionId=None,
            isSuccess=None,
            isPending=None,
            isError=None,
            card3ds=None,
            redirectUrl=None,
            messageCode=None,
            messageMerchant=None,
            messageCustomer=None,
            amount=None,
            effectiveInterestRate=None,
            currency=None,
            returnUrl=None,
            date=None,
            customerId=None,
            paymentId=None,
            basketId=None,
            metadataId=None,
            payPageId=None,
            linkPayId=None,
            typeId=None,
            orderId=None,
            invoiceId=None,
            paymentReference=None,
            processing=None,
            **kwargs
    ):
        """Create a new PaymentResponse.

        :param transactionId: Id of this charge transaction
        :type transactionId: str
        :param isSuccess: (optional)
        :type isSuccess: bool
        :param isPending: (optional)
        :type isPending: bool
        :param isError: (optional)
        :type isError: bool
        :param card3ds: (optional) Indicate a 3ds transaction (card payment type only).
        :type card3ds: bool
        :param redirectUrl: (optional)  Some payment methods require the customer
            to leave the merchant application.
            This URL is used to bring the customer back to your application.
        :type redirectUrl: str
        :param messageCode: (optional) Response message of payment Core. Code of the message.
        :type messageCode: str
        :param messageMerchant: (optional) Response message of payment Core. Message for merchant.
        :type messageMerchant: str
        :param messageCustomer: (optional) Response message of payment Core. Message for customer.
        :type messageCustomer: str
        :param amount: (optional) The amount to be authorized on the specified account.
            The amount is rounded depending on the respective currency.
        :type amount: float
        :param effectiveInterestRate: (optional) Only valid for Installment method:
            The affected installment rated. Required in case of Installment method.
        :type effectiveInterestRate: str
        :param currency: (optional) ISO currency code.
        :type currency: str
        :param returnUrl: (optional) If customer's confirmation is required, a redirect URL will be return.
            Customer needs to be redirected to this URL and proceed the confirmation.
        :type returnUrl: str
        :param date: (optional) Timestamp of this transaction.
        :type date: datetime.datetime

        Resources
        :param customerId: (optional) Customer id used for this transaction.
        :type customerId: str
        :param paymentId: (optional) Id of the payment.
        :type paymentId: str
        :param basketId: (optional) Basket ID used for this transaction.
        :type basketId: str
        :param metadataId: (optional) Meta data ID used for this transaction.
        :type metadataId: str
        :param payPageId: (optional) Payment Page Id related to this payment.
        :type payPageId: str
        :param linkPayId: (optional)
        :type linkPayId: str
        :param typeId: (optional) Id of the types Resource that is to be used for this transaction.
        :type typeId: str

        :param orderId: (optional) Order id that identifies the payment on merchant side.
        :type orderId: str
        :param invoiceId: (optional) invoice id that is assigned to the payment on merchant side.
        :type invoiceId: str
        :param paymentReference: (optional) Transaction description.
        :type paymentReference: str
        :param processing: (optional)
        :type processing: PaymentResponseMetadata
        """
        super().__init__(**kwargs)
        self.transactionId = transactionId  # type:str
        self.isSuccess = isSuccess  # type:bool
        self.isPending = isPending  # type:bool
        self.isError = isError  # type:bool
        self.card3ds = card3ds  # type:bool
        self.redirectUrl = redirectUrl  # type:str
        self.messageCode = messageCode  # type:str
        self.messageMerchant = messageMerchant  # type:str
        self.messageCustomer = messageCustomer  # type:str
        self.amount = amount  # type:float
        self.effectiveInterestRate = effectiveInterestRate  # type:str
        self.currency = currency  # type:str
        self.returnUrl = returnUrl  # type:str
        self.date = date  # type:datetime.datetime
        self.customerId = customerId  # type:str
        self.paymentId = paymentId  # type:str
        self.basketId = basketId  # type:str
        self.metadataId = metadataId  # type:str
        self.payPageId = payPageId  # type:str
        self.linkPayId = linkPayId  # type:str
        self.typeId = typeId  # type:str
        self.orderId = orderId  # type:str
        self.invoiceId = invoiceId  # type:str
        self.paymentReference = paymentReference  # type:str
        self.processing = processing  # type:PaymentResponseMetadata

    def serialize(self):
        raise NotImplementedError("No serialisation for response models.")

    @classmethod
    def fromDict(cls, data: dict, client: "UnzerClient") -> t.Self:
        data = data.copy()
        data["transactionId"] = data["id"]
        data["isSuccess"] = parseBool(data["isSuccess"])
        data["isPending"] = parseBool(data["isPending"])
        data["isError"] = parseBool(data["isError"])
        data["card3ds"] = parseBool(data["card3ds"]) if "card3ds" in data else None
        data["amount"] = float(data["amount"])
        data["date"] = parseDateTime(data["date"])
        data["processing"] = PaymentResponseMetadata.fromDict(data["processing"])
        # Message
        if not data["message"]:
            data["message"] = {}
        data["messageCode"] = data["message"].get("code")
        data["messageMerchant"] = data["message"].get("merchant")
        data["messageCustomer"] = data["message"].get("customer")
        # Resources
        data["customerId"] = data["resources"].get("customerId") or None
        data["paymentId"] = data["resources"].get("paymentId") or None
        data["basketId"] = data["resources"].get("basketId") or None
        data["metadataId"] = data["resources"].get("metadataId") or None
        data["payPageId"] = data["resources"].get("payPageId") or None
        data["linkPayId"] = data["resources"].get("linkPayId") or None
        data["typeId"] = data["resources"].get("typeId") or None
        return cls(**data, client=client)

    def charge(self, amount: float) -> "PaymentResponse":
        req_kwargs = self.__dict__.copy()
        paymentTypeName = PaymentGetResponse.getPaymentTypeFromTypeId(self.typeId)
        req_kwargs["paymentType"] = PaymentType.construct(paymentTypeName)(self.typeId)
        req_kwargs["amount"] = amount
        logging.debug("req_kwargs = %r", req_kwargs)
        req = PaymentRequest(**req_kwargs)
        return self._client.charge(req)


class PaymentResponseMetadata(BaseModel):
    def __init__(
            self,
            creatorId=None,
            identification=None,
            iban=None,
            bic=None,
            bank=None,
            externalOrderId=None,
            zgReferenceId=None,
            traceId=None,
            basketId=None,
            uniqueId=None,
            shortId=None,
            descriptor=None,
            holder=None,
            PDFLink=None,
            paypalBuyerId=None,
            threeDsEci=None,
            participantId=None,
            **kwargs
    ):
        """Create a new PaymentResponseMetadata.

        :param creatorId: (optional) String This value returns your creditor id.
        :type creatorId: str
        :param identification: (optional) String This value returns the descriptor for invoice and prepayment.
        :type identification: str
        :param iban: (optional) String Iban of the merchant for prepayment or invoice.
            In the case of a direct debit, this value contains the customer Iban.
        :type iban: str
        :param bic: (optional) String Bic of the merchant for prepayment or invoice.
            In the case of a direct debit, this value contains the customer Bic.
        :type bic: str
        :param bank: (optional)
            Bank of the merchant for prepayment or invoice.
            In the case of a direct debit, this value contains the customer Bank.
        :type bank: str
        :param externalOrderId: (optional) String External Order Id of installment transaction
            e.g: Hirepurchase, Installment-Secured.
        :type externalOrderId: str
        :param zgReferenceId: (optional) String Reference Id of installment transaction
            e.g: Hirepurchase, Installment-Secured.
        :type zgReferenceId: str
        :param traceId: (optional)
        :type traceId: str
        :param basketId: (optional) String Basket ID used for this transaction.
        :type basketId: str
        :param uniqueId: (optional) String Unique id of the payment system used.
        :type uniqueId: str
        :param shortId: (optional) String User-friendly reference id of the payment system.
        :type shortId: str
        :param descriptor: (optional) String Descriptor of the merchant for prepayment or invoice..
        :type descriptor: str
        :param holder: (optional) String Holder of the merchant for prepayment or invoice.
            In the case of a direct debit, this value contains the customer holder.
        :type holder: str
        :param PDFLink: (optional) String PDFLink of installment transaction
            e.g: Hirepurchase, Installment-Secured.
        :type PDFLink: str
        :param paypalBuyerId: (optional) String Id of buyer for Paypal transaction.
        :type paypalBuyerId: str
        :param threeDsEci: (optional) String 3dsEci flag from Payment Core.
        :type threeDsEci: str
        :param participantId: String Only valid for marketplace payment:
            Channel Id(s) of marketplace's participant(s).
        :type participantId: str
        """
        super().__init__(**kwargs)
        self.creatorId = creatorId  # type:str
        self.identification = identification  # type:str
        self.iban = iban  # type:str
        self.bic = bic  # type:str
        self.bank = bank  # type:str
        self.externalOrderId = externalOrderId  # type:str
        self.zgReferenceId = zgReferenceId  # type:str
        self.traceId = traceId  # type:str
        self.basketId = basketId  # type:str
        self.uniqueId = uniqueId  # type:str
        self.shortId = shortId  # type:str
        self.descriptor = descriptor  # type:str
        self.holder = holder  # type:str
        self.PDFLink = PDFLink  # type:str
        self.paypalBuyerId = paypalBuyerId  # type:str
        self.threeDsEci = threeDsEci  # type:str
        self.participantId = participantId  # type:str

    def serialize(self):
        raise NotImplementedError("No serialisation for response models.")

    @classmethod
    def fromDict(cls, data):
        data = data.copy()
        # Nobody, really nobody starts identifier with a digit. Unzer: here you have the 3dsEci flag
        data["threeDsEci"] = data["3dsEci"] if "3dsEci" in data else None
        return cls(**data)


from unzer.model.payment_type.abstract_paymenttype import PaymentType  # noqa: Avoid circular imports



from unzer.model.payment import PaymentMethodTypes, PaymentTypes
from .abstract_paymenttype import PaymentType


class Sofort(PaymentType):
    method = PaymentTypes.SOFORT
    method_name = PaymentMethodTypes.SOFORT



import abc
import logging
import typing as t

from ..base import BaseModel

if t.TYPE_CHECKING:
    from ..payment import PaymentTypes, PaymentMethodTypes  # noqa

logger = logging.getLogger("unzer-sdk").getChild(__name__)


class PaymentType(BaseModel):
    @property
    @abc.abstractmethod
    def method(self) -> "PaymentTypes":
        """Hold the type."""
        pass

    @property
    @abc.abstractmethod
    def method_name(self) -> "PaymentMethodTypes":
        """Hold the full payment name."""
        pass

    # TODO: Rename method and method_name (Unzer himself has no clear concept or consistent naming for this either)

    def __init__(
            self,
            key: str = None,
            **kwargs
    ):
        """Create a new paymentType ressource.

        :param key: (optional) (original: id) ID for this payment type
        """
        super().__init__(**kwargs)
        self.key: str = key

    def serialize(self) -> dict:
        return {}

    @classmethod
    def fromDict(cls, data: dict) -> t.Self:
        data = data.copy()
        data["key"] = data["id"]
        return cls(**data)

    @classmethod
    def get_subclasses(cls) -> t.Type[t.Self]:
        for subclass in cls.__subclasses__():
            yield from subclass.get_subclasses()
            yield subclass

    @classmethod
    def construct(cls, method: "PaymentTypes") -> t.Type["PaymentType"]:
        for subclass in PaymentType.get_subclasses():
            if subclass.method == method:
                return subclass
        logger.warning(f"Creating not existing PaymentType for method {method} on the fly")
        sub_cls = type(str(method).title(), (cls,), {"method": method, "method_name": "N/A"})
        return sub_cls  # noqa

    def get_configuration(self) -> dict:
        if self._client is None:
            raise IOError(f"PaymentType {type(self).__name__} was not initialized with client instance")
        key_pair_types = self._client.getKeyPairTypes()
        logger.debug(f"key_pair_types: {key_pair_types}")
        for payment_type in key_pair_types["paymentTypes"]:
            if payment_type["type"] == self.method_name.value:
                return payment_type
        raise LookupError(f"PaymentType {self.method_name} is not configured in the keypair")

    # TODO: Without caching this isn't the best solution --> better implement in the KeyPairTypeModel
    def get_channel_id(self) -> str:
        return self.get_configuration()["supports"][0]["channel"]

    def get_brands(self) -> list[str]:
        return self.get_configuration()["supports"][0]["brands"]



from .base import BaseModel


class BasketItem(BaseModel):
    def __init__(
            self,
            basketItemReferenceId=None,
            unit=None,
            quantity=None,
            amountDiscount=None,
            vat=None,
            amountGross=None,
            amountVat=None,
            amountPerUnit=None,
            amountNet=None,
            title=None,
            subTitle=None,
            imageUrl=None,
            participantId=None,
            kind=None,
            **kwargs
    ):
        """Create a new BasketItem.

        :param basketItemReferenceId: (optional) Unique basket item reference ID (within the basket)
        :type basketItemReferenceId: str
        :param unit: (optional) Unit description of the item e.g. &quot;pc&quot;
        :type unit: str
        :param quantity: Integer Quantity of the basket item format: int32
        :type quantity: int
        :param amountDiscount: (optional) Discount amount for the basket item
            (multiplied by the :attr:`quantity`) format: float
        :type amountDiscount: float
        :param vat: (optional) Integer Vat value for the basket item in percent (0-100) format: int32
        :type vat: int
        :param amountGross: (optional) Gross amount (= amountNet + amountVat) in the specified currency.
            Equals amountNet if vat value is 0 format: float
        :type amountGross: float
        :param amountVat: (optional) Vat amount. Equals 0 if vat value is 0.
            Should equal the :attr:`vat` multiplied by :attr:`amountNet` for each basket item. format: float
        :type amountVat: float
        :param amountPerUnit: NET amount per unit format: float
        :type amountPerUnit: float
        :param amountNet: (optional) Net amount. Equals amountGross if vat value is 0. format: float
        :type amountNet: str
        :param title: Title of the basket item (max. 255)
        :type title: str
        :param subTitle: (optional) The defined subTitle which is displayed on our Payment Page later on
        :type subTitle: str
        :param imageUrl: (optional) The defined imageUrl for the related basketItem
            and will be displayed on our Payment Page
        :type imageUrl: str
        :param participantId: (optional) Only valid for marketplace payment:
            Channel Id(s) of marketplace's participant(s).
        :type participantId: str
        :param kind: (original: type) (optional)
        :type kind: str
        """
        super().__init__(**kwargs)
        self.basketItemReferenceId = basketItemReferenceId
        self.unit = unit
        self.quantity = quantity
        self.amountDiscount = amountDiscount
        self.vat = vat
        self.amountGross = amountGross
        self.amountVat = amountVat
        self.amountPerUnit = amountPerUnit
        self.amountNet = amountNet
        self.title = title
        self.subTitle = subTitle
        self.imageUrl = imageUrl
        self.participantId = participantId
        self.kind = kind

    def serialize(self):
        return {
            "basketItemReferenceId": self.getString(self.basketItemReferenceId),
            "unit": self.getString(self.unit),
            "quantity": self.quantity,
            "amountDiscount": self.amountDiscount,
            "vat": self.vat,
            "amountGross": self.amountGross,
            "amountVat": self.amountVat,
            "amountPerUnit": self.amountPerUnit,
            "amountNet": self.amountNet,
            "title": self.getString(self.title),
            "subTitle": self.getString(self.subTitle),
            "imageUrl": self.getString(self.imageUrl),
            "participantId": self.getString(self.participantId),
            "type": self.getString(self.kind),
        }

    @classmethod
    def fromDict(cls, data):
        data = data.copy()
        data["kind"] = data["type"]
        data["amountGross"] = float(data["amountGross"])
        data["amountVat"] = float(data["amountVat"])
        data["amountPerUnit"] = float(data["amountPerUnit"])
        data["amountNet"] = float(data["amountNet"])
        data["quantity"] = int(data["quantity"])
        data["vat"] = float(data["vat"])
        return cls(**data)
